//$AbstractBorderItem,BorderItem,WindowBorderItem,ShadowBorderItem$

#include "BorderItems.h"
#include "String.h"

const int cBorderTitleGap= 4;
const int cHTitleGap    = 10;
const int cBorder       = 9,
	  cTitleGap     = 5;        // gap between titlebar and contents


static short StretchBoxNEBits[]= {
#   include "images/StretchBoxNE.image"
};

static short StretchBoxNWBits[]= {
#   include "images/StretchBoxNW.image"
};

static short StretchBoxSEBits[]= {
#   include "images/StretchBoxSE.image"
};

static short StretchBoxSWBits[]= {
#   include "images/StretchBoxSW.image"
};

static StaticBitmap StretchBoxNW(8, StretchBoxNWBits),
		    StretchBoxNE(8, StretchBoxNEBits),
		    StretchBoxSW(8, StretchBoxSWBits),
		    StretchBoxSE(8, StretchBoxSEBits);

//---- AbstractBorderItem ------------------------------------------------------

AbstractMetaImpl(AbstractBorderItem, (I_P(interiorOffset), I_P(extentDiff), I_P(border),
    I_O(interior), I_O(titleBar), I_S(lineWidth), I_S(titleGap), I_S(shadowWidth)));

AbstractBorderItem::AbstractBorderItem(VObject *ti, VObject *in, int lw, int id)
							: (id, in, ti, 0)
{
    interior= in;
    titleBar= ti;
    lineWidth= lw;
}

AbstractBorderItem::AbstractBorderItem(char *ti, VObject *in, int lw= 1, int id)
				    : (id, in, titleBar= new TextItem(ti), 0)
{
    interior= in;
    lineWidth= lw;
}

Metric AbstractBorderItem::GetMinSize()
{
    Point e= interior->GetMinSize().extent;
    
    interiorOffset= border;
    extentDiff= 2*interiorOffset + shadowWidth;
    
    if (titleBar) {
	if (titleBar->IsKindOf(CompositeVObject))
	    ((CompositeVObject*)titleBar)->SetModified();
	Metric tm= titleBar->GetMinSize();
	tm.extent.x+= cHTitleGap*2;
	tm.extent.y+= titleGap;
	e.x= max(e.x, tm.extent.x);
	interiorOffset.y+= tm.extent.y;
	extentDiff.y+= tm.extent.y;
    }
    return e+extentDiff;
}

void AbstractBorderItem::SetOrigin(Point at)
{
    VObject::SetOrigin(at);
    interior->SetOrigin(at+interiorOffset);
    if (titleBar) {
	at.y+= border.y;
	if (titleBar->TestFlag(eVObjHFixed))
	    at.x+= (Width() - shadowWidth - titleBar->Width())/2;
	else
	    at.x+= border.x+cHTitleGap;
	titleBar->SetOrigin(at);
    }
}

void AbstractBorderItem::SetExtent(Point e)
{
    VObject::SetExtent(e);
    e-= extentDiff;
    if (titleBar) {
	if (titleBar->TestFlag(eVObjHFixed))
	    titleBar->CalcExtent();
	else
	    titleBar->SetExtent(Point(e.x-cHTitleGap*2, interiorOffset.y-border.y));
    }
    interior->SetExtent(e);
}

ostream& AbstractBorderItem::PrintOn(ostream &s)
{
    CompositeVObject::PrintOn(s);
    return s << border SP << lineWidth SP << interiorOffset SP << extentDiff SP
						<< interior SP << titleBar SP;
}

istream& AbstractBorderItem::ReadFrom(istream &s)
{
    CompositeVObject::ReadFrom(s);
    return s >> border >> lineWidth >> interiorOffset >> extentDiff
						       >> interior >> titleBar;
}

void AbstractBorderItem::SetTitle(char *itsTitle, bool redraw)
{
    if (titleBar == 0)
	return;
	
    // ugly hack
    VObject *tit= titleBar->FindItem(cIdWindowTitle);
    if (tit == 0)
	tit= titleBar;
    if (tit && tit->IsKindOf(TextItem)) {
	((TextItem*)tit)->SetString(itsTitle);
	tit->CalcExtent();
	if (titleBar->IsKindOf(CompositeVObject))
	    ((CompositeVObject*)titleBar)->SetModified();
	if (redraw && IsOpen()) {
	    Rectangle oldr= titleBar->contentRect;
	    titleBar->CalcExtent();
	    titleBar->SetExtent(oldr.extent);
	    titleBar->SetOrigin(oldr.origin);
	    InvalidateRect(oldr);
	}
    }
}

char *AbstractBorderItem::GetTitle()
{
    if (titleBar == 0)
	return 0;
	
    VObject *tit= titleBar->FindItem(cIdWindowTitle);
    if (tit == 0)
	tit= titleBar;
    if (tit && tit->IsKindOf(TextItem))
	return tit->AsString();
    return 0;
}

void AbstractBorderItem::InspectorId(char *buf, int sz)
{
    if (titleBar == 0)
	return;
	
    VObject *tit= titleBar->FindItem(cIdWindowTitle);
    if (tit && tit->IsKindOf(TextItem)) 
	strn0cpy(buf, ((TextItem*)tit)->AsString(), sz);
}

//---- ShadowBorderItem --------------------------------------------------------

MetaImpl(ShadowBorderItem, I_S(shadowWidth));

ShadowBorderItem::ShadowBorderItem(VObject *title, VObject *inner, int shadow)
						: (title, inner, 0, cIdNone)
{
    shadowWidth= shadow;
    border= gPoint2;
}

ShadowBorderItem::ShadowBorderItem(char *title, VObject *inner, int shadow)
						: (title, inner, 0, cIdNone)
{
    shadowWidth= shadow;
    border= gPoint2;
}

ShadowBorderItem::ShadowBorderItem(VObject *inner, int shadow)
					    : ((VObject*)0, inner, 0, cIdNone)
{
    shadowWidth= shadow;
    border= gPoint2;
}

void ShadowBorderItem::DrawBackground(Rectangle r)
{
    Rectangle rr= contentRect;

    rr.extent-= shadowWidth;
    GrPaintRect(rr, ePatWhite);
    if (! interior->contentRect.ContainsRect(r)) {
	if (shadowWidth) {
	    GrPattern pat= shadowWidth == 1 ? ePatBlack : ePatGrey50;
	    GrPaintRect(Rectangle(rr.origin.x+rr.extent.x, rr.origin.y+shadowWidth, shadowWidth, rr.extent.y), pat);
	    GrPaintRect(Rectangle(rr.origin.x+shadowWidth, rr.origin.y+rr.extent.y, rr.extent.x, shadowWidth), pat);
	}
	GrSetPenNormal();
	GrStrokeRect(rr);
    }
}

void ShadowBorderItem::DrawForeground(Rectangle)
{
    if (titleBar) {
	Rectangle r= contentRect;
	r.extent-= shadowWidth;
	r.extent.y= titleBar->Height();
	GrInvertRect(r.Inset(2));
    }
}

//---- WindowBorderItem --------------------------------------------------------------

MetaImpl(WindowBorderItem, I_B(fixed));

WindowBorderItem::WindowBorderItem(bool fix, VObject *title, VObject *inner)
						: (title, inner, 1, cIdNone)
{
    fixed= fix;
    border= cBorder;
    titleGap= cTitleGap;
}

void WindowBorderItem::DrawBackground(Rectangle r)
{
    GrEraseRect(r);
    if (! interior->contentRect.ContainsRect(r)) {
	GrSetPenNormal();
	GrSetPenSize(2);
	GrStrokeRect(contentRect);
	GrSetPenSize(1);
	if (! fixed) {
	    GrPaintBitMap(Rectangle(contentRect.NW()+Point(2), 8),
						    &StretchBoxNW, ePatBlack);
	    GrPaintBitMap(Rectangle(contentRect.NE()+Point(-9,2), 8),
						    &StretchBoxNE, ePatBlack);
	    GrPaintBitMap(Rectangle(contentRect.SE()+Point(-9,-9), 8),
						    &StretchBoxSE, ePatBlack);
	    GrPaintBitMap(Rectangle(contentRect.SW()+Point(2,-9), 8),
						    &StretchBoxSW, ePatBlack);
	}
    }
}

Command *WindowBorderItem::DoLeftButtonDownCommand(Point p, Token t, int cl)
{
    if (Rectangle(contentRect.NW()+Point(2), 8).ContainsPoint(p)
	    || Rectangle(contentRect.NE()+Point(-9,2), 8).ContainsPoint(p)
	    || Rectangle(contentRect.SE()+Point(-9,-9), 8).ContainsPoint(p)
	    || Rectangle(contentRect.SW()+Point(2,-9), 8).ContainsPoint(p)) {
	t.Flags|= eFlgCntlKey;
	return EvtHandler::DoMiddleButtonDownCommand(p, t, cl);
    }
    return EvtHandler::DoLeftButtonDownCommand(p, t, cl);
}

ostream& WindowBorderItem::PrintOn(ostream &s)
{
    AbstractBorderItem::PrintOn(s);
    return s << fixed SP;
}

istream& WindowBorderItem::ReadFrom(istream &s)
{
    AbstractBorderItem::ReadFrom(s);
    return s >> Enum(fixed);
}

//---- BorderItem --------------------------------------------------------------

MetaImpl(BorderItem, I_I(titleAlign));

BorderItem::BorderItem(VObject *ip, Point b, int lw, VObjAlign a)
						: ((VObject*)0, ip, lw, cIdNone)
{
    border= b;
    titleAlign= a;
}

BorderItem::BorderItem(VObject *ti, VObject *ip, Point b, int lw, VObjAlign a)
							: (ti, ip, lw, cIdNone)
{
    border= b;
    titleAlign= a;
}

BorderItem::BorderItem(char *ti, VObject *ip, Point b, int lw, VObjAlign a)
							: (ti, ip, lw, cIdNone)
{
    border= b;
    titleAlign= a;
}

BorderItem::BorderItem(int id, char *ti, VObject* ip, Point b, int lw, VObjAlign a)
							: (ti, ip, lw, id)
{
    border= b;
    titleAlign= a;
}

void BorderItem::DrawBackground(Rectangle r)
{
    if (! interior->contentRect.ContainsRect(r)) {
	GrSetPenNormal();
	if (lineWidth >= 0) {
	    GrSetPenSize(lineWidth);
	    if (! Enabled())
		GrSetPenPattern(ePatGrey50);
	}
	if (titleBar) {
	    if (lineWidth >= 0) {
		int h= titleBar->Height()/2;
		Rectangle rr= contentRect;
		rr.origin.y+= h;
		rr.extent.y-= h;
		GrStrokeRect(rr);
	    }
	    GrEraseRect(titleBar->contentRect);
	} else if (lineWidth >= 0)
	    GrStrokeRect(contentRect);
	GrSetPenNormal();
    }
}

void BorderItem::SetOrigin(Point at)
{
    int lw= lineWidth >= 0 ? lineWidth : 0;  
    VObject::SetOrigin(at);
    if (titleBar) {
	int th= titleBar->Height();
	int tgap= border.x+lw+cBorderTitleGap;
	titleBar->Align(Point(at.x+tgap, at.y), 
		Metric(Width()-2*tgap, th),(titleAlign & eVObjH) | eVObjVTop);
	int d= th - (border.y+lw);
	if (d > 0)
	    at.y+= d;
    }
    interior->SetOrigin(at+border+Point(lw));
}

Metric BorderItem::GetMinSize()
{
    int lw= lineWidth >= 0 ? lineWidth : 0;  
    Metric m= interior->GetMinSize();
    
    if (titleBar) {
	Metric tm= titleBar->GetMinSize();
	m.extent.x= max(m.extent.x, tm.extent.x+2*cBorderTitleGap);
	int d= tm.extent.y - (border.y+lw);
	if (d > 0) {
	    m.extent.y+= d;
	    m.base+= d;
	}
    }
    return m.Expand(border+Point(lw));
}

void BorderItem::SetExtent(Point e)
{
    int d= 0, lw= lineWidth >= 0 ? lineWidth : 0;  

    VObject::SetExtent(e);
    if (titleBar) {
	titleBar->CalcExtent();
	d= titleBar->Height() - (border.y+lw);
	if (d > 0)
	    e.y-= d;
    }
    interior->SetExtent(e-2*(border+Point(lw)));
}

int BorderItem::Base()
{
    int d= 0, lw= lineWidth >= 0 ? lineWidth : 0;
      
    if (titleBar) {
	d= titleBar->Height() - (border.y+lw);
	if (d < 0)
	    d= 0;
    }
    return interior->Base()+d+border.y+lw;
}

ostream& BorderItem::PrintOn(ostream &s)
{
    AbstractBorderItem::PrintOn(s);
    return s << titleAlign SP;
}

istream& BorderItem::ReadFrom(istream &s)
{
    AbstractBorderItem::ReadFrom(s);
    return s >> Enum(titleAlign);
}
